
import java.lang.InstantiationException;


public class Transform {
	
	/**
	 * auxilary function for world2camera
	 * @param p 
	 * @return the 4x4 transformation matrix from world to camera referency
	 * @throws Exception if the up vector or the Z vector is 0
	 */
	private static double[][] matWorld2Cam(Parms p) throws Exception{
		double trans[][]= new double [4][4];
		double chBase[][]= new double [4][4];
		double[] upVector = p.getCameraUp();
		double[] interestPoint = p.getCameraLookAt();
		double[] viewPoint = p.getCameraPosition();
		double normy = MatVec.norm3(upVector);
		double z[] = new double[3];
		
		/** 
		 * Defintion of the 4x4 translation matrix
		 */
		for (int i = 0; i < trans.length; i++) {
			trans[i][i] = 1.0;
		}
		for (int i = 0; i < viewPoint.length; i++) {
			trans[i][3] = -viewPoint[i];
		}

		//System.out.println("Matrice de translation");
		//MatVec.printMat(trans);

		
		/** 
		 * Definition of the rotation matrix (put in a 4X4 matrix, in
		 * order to do the product later
		 */
		chBase[3][3]= 1.0;

		//z = MatVec.sousVec3(interestPoint, viewPoint);
		for (int i = 0; i < z.length; i++) {
			z[i] = interestPoint[i] - viewPoint[i];
		}

		if ( normy == 0 ){ 
			throw new InstantiationException("up vecteur est nul!!");
		}
		
		// y dans le sens du up vecteur
		for (int i = 0; i < 3; i++) {
			chBase[1][i]=upVector[i]/normy;
		}
		// z du viewPoint vers le interestPoint
		{
			double normz;
			
			normz = MatVec.norm3(z);
			if ( normz == 0 ){ 
				throw new InstantiationException("Z vecteur est nul!!");
			}
			for (int i = 0; i < z.length; i++) {
				chBase[2][i] = z[i]/normz;
			}
		}
		// x comme produit vectoriel de y ^ z
		{ 
			double[] x = new double[3];
			double normx;
			
			x = MatVec.prodVec(chBase[1], chBase[2]);
			normx = MatVec.norm3(x);
			for (int j = 0; j < x.length; j++) {
				chBase[0][j] = x[j]/normx;
			}
		}
		
		{// si y non perpendiculaire a z alors y=z^x
			double epsilon = 0.000001;
			
			if (Math.abs(MatVec.dotProd3(chBase[1],chBase[2])) > epsilon)
			{
				double[] y = new double[3];
				double normY;
				
				y = MatVec.prodVec(chBase[2], chBase[0]);
				normY = MatVec.norm3(y);
				for (int j = 0; j < y.length; j++) {
					chBase[1][j] = y[j]/normY;
				}	
				// System.out.println("ouhouh");
			}	
				
		}
		//System.out.println("Matrice de changement de repere");
		//MatVec.printMat(MatVec.multMat44Mat(chBase, trans));
	
		return MatVec.multMat44Mat(chBase, trans);	
	}
	
	/**
	 * applies the transformation to an array of points
	 * @param oldVertex array of points
	 * @param p 
	 * @return newVertex contains the coordinates pf the same poits in the camera referency
	 * @throws Exception
	 */
	public static double[] world2Camera (double oldVertex[], Parms p) throws Exception 
	{
		//contient les vecteurs ligne de la base camera dans la base monde
		double newVertex[] = new double[oldVertex.length];
		double[][] chBase = matWorld2Cam(p) ;
		
		/**
		 * Application du changement de base a tous les sommets avec la multiplication matricielle
		 */
		{
			double vertex[] = new double[4];
			
			for (int j = 0; j < oldVertex.length; j=j+3) {
				for (int i = 0; i < 3; i++) {		
					vertex[i] = oldVertex[j+i];
				}
				vertex[3]=1.0;
				vertex = MatVec.multMat44Vec(chBase, vertex);
				for (int i = 0; i < 3; i++) {
					newVertex[j+i] = vertex[i];	
					}
			}		
		}

		return newVertex;
	}	
	
	/**
	 * projection in the plane z=focal. Z coordinate of the point is kept of Z-buffering
	 * @param verticesW
	 * @return projected vertices with coord (x_p, y_p, old_z) where old_z is the original z
	 * coordinate
	 */
	public static double[] projection( double[] verticesW, Parms p ){
		double[] projVertices = new double[verticesW.length];
		
		for (int i = 0; i < verticesW.length; i=i+3) {
			double zsurzf = p.getCameraFocal()/verticesW[i+2];
			projVertices[i]=verticesW[i]*zsurzf;
			projVertices[i+1]=verticesW[i+1]*zsurzf;
			projVertices[i+2]=verticesW[i+2];
		}
		
		return projVertices;
	}
	
	/**
	 * computes image coordinates of vertices assumed in the projection plane
	 * @param verticesC
	 * @param p
	 * @return images coordinates + original Z_coordinate
	 */
	public static double[] camera2Raw2D(double[] verticesC, Parms p){
		double xLim = p.getCameraXLimit();
		int h = p.getScreenH();
		int w = p.getScreenW();
		double yLim = xLim * h / w ;
		double[] verticesR2D = new double[verticesC.length];
		
		for (int i = 0; i < verticesC.length; i=i+3) {
			verticesR2D[i] = -(verticesC[i] - xLim)* w / (2*xLim);
			verticesR2D[i+1] = -(verticesC[i+1] - yLim)* h / (2*yLim);
			verticesR2D[i+2] = verticesC[i+2];
		}
		
		return verticesR2D;
	}
}
